@@ -2,8 +2,10 @@ use std::str;
 
 use nom::branch::alt;
 use nom::bytes::complete::{tag, take_till};
-use nom::character::complete::char;
-use nom::combinator::{complete, consumed, cut, eof, fail, map, not, opt, peek, recognize, value};
+use nom::character::complete::{anychar, char};
+use nom::combinator::{
+    complete, consumed, cut, eof, fail, map, map_opt, not, opt, peek, recognize, value,
+};
 use nom::multi::{many0, separated_list0, separated_list1};
 use nom::sequence::{delimited, pair, preceded, tuple};
 
@@ -35,13 +37,38 @@ pub enum Node<'a> {
 }
 
 impl<'a> Node<'a> {
-    pub(super) fn many(i: &'a str, s: &State<'_>) -> ParseResult<'a, Vec<Self>> {
-        complete(many0(alt((
+    pub(super) fn parse_template(i: &'a str, s: &State<'_>) -> ParseResult<'a, Vec<Self>> {
+        let (i, result) = match complete(|i| Self::many(i, s))(i) {
+            Ok((i, result)) => (i, result),
+            Err(err) => {
+                if let nom::Err::Error(err) | nom::Err::Failure(err) = &err {
+                    if err.message.is_none() {
+                        opt(|i| unexpected_tag(i, s))(err.input)?;
+                    }
+                }
+                return Err(err);
+            }
+        };
+        let (i, _) = opt(|i| unexpected_tag(i, s))(i)?;
+        let (i, is_eof) = opt(eof)(i)?;
+        if is_eof.is_none() {
+            return Err(nom::Err::Failure(ErrorContext::new(
+                "cannot parse entire template\n\
+                 you should never encounter this error\n\
+                 please report this error to <https://github.com/rinja-rs/rinja/issues>",
+                i,
+            )));
+        }
+        Ok((i, result))
+    }
+
+    fn many(i: &'a str, s: &State<'_>) -> ParseResult<'a, Vec<Self>> {
+        many0(alt((
             map(|i| Lit::parse(i, s), Self::Lit),
             map(|i| Comment::parse(i, s), Self::Comment),
             |i| Self::expr(i, s),
             |i| Self::parse(i, s),
-        ))))(i)
+        )))(i)
     }
 
     fn parse(i: &'a str, s: &State<'_>) -> ParseResult<'a, Self> {
@@ -77,15 +104,15 @@ impl<'a> Node<'a> {
             "break" => |i, s| Self::r#break(i, s),
             "continue" => |i, s| Self::r#continue(i, s),
             "filter" => |i, s| wrap(Self::FilterBlock, FilterBlock::parse(i, s)),
-            _ => return fail(i),
+            _ => return fail(start),
         };
 
         let (i, node) = s.nest(j, |i| func(i, s))?;
 
-        let (i, closed) = cut(alt((
-            value(true, |i| s.tag_block_end(i)),
-            value(false, ws(eof)),
-        )))(i)?;
+        let (i, closed) = cut_node(
+            None,
+            alt((value(true, |i| s.tag_block_end(i)), value(false, ws(eof)))),
+        )(i)?;
         match closed {
             true => Ok((i, node)),
             false => Err(ErrorContext::unclosed("block", s.syntax.block_end, start).into()),
@@ -128,16 +155,22 @@ impl<'a> Node<'a> {
         let start = i;
         let (i, (pws, expr)) = preceded(
             |i| s.tag_expr_start(i),
-            cut(pair(
-                opt(Whitespace::parse),
-                ws(|i| Expr::parse(i, s.level.get())),
-            )),
+            cut_node(
+                None,
+                pair(
+                    opt(Whitespace::parse),
+                    ws(|i| Expr::parse(i, s.level.get())),
+                ),
+            ),
         )(i)?;
 
-        let (i, (nws, closed)) = cut(pair(
-            opt(Whitespace::parse),
-            alt((value(true, |i| s.tag_expr_end(i)), value(false, ws(eof)))),
-        ))(i)?;
+        let (i, (nws, closed)) = cut_node(
+            None,
+            pair(
+                opt(Whitespace::parse),
+                alt((value(true, |i| s.tag_expr_end(i)), value(false, ws(eof)))),
+            ),
+        )(i)?;
         match closed {
             true => Ok((i, Self::Expr(Ws(pws, nws), expr))),
             false => Err(ErrorContext::unclosed("expression", s.syntax.expr_end, start).into()),
@@ -167,6 +200,48 @@ impl<'a> Node<'a> {
     }
 }
 
+fn cut_node<'a, O>(
+    kind: Option<&'static str>,
+    inner: impl FnMut(&'a str) -> ParseResult<'a, O>,
+) -> impl FnMut(&'a str) -> ParseResult<'a, O> {
+    let mut inner = cut(inner);
+    move |i: &'a str| {
+        let result = inner(i);
+        if let Err(nom::Err::Failure(err) | nom::Err::Error(err)) = &result {
+            if err.message.is_none() {
+                opt(|i| unexpected_raw_tag(kind, i))(err.input)?;
+            }
+        }
+        result
+    }
+}
+
+fn unexpected_tag<'a>(i: &'a str, s: &State<'_>) -> ParseResult<'a, ()> {
+    value(
+        (),
+        tuple((
+            |i| s.tag_block_start(i),
+            opt(Whitespace::parse),
+            |i| unexpected_raw_tag(None, i),
+        )),
+    )(i)
+}
+
+fn unexpected_raw_tag<'a>(kind: Option<&'static str>, i: &'a str) -> ParseResult<'a, ()> {
+    let (_, tag) = ws(identifier)(i)?;
+    let msg = match tag {
+        "end" | "elif" | "else" | "when" => match kind {
+            Some(kind) => {
+                format!("node `{tag}` was not expected in the current context: `{kind}` block")
+            }
+            None => format!("node `{tag}` was not expected in the current context"),
+        },
+        tag if tag.starts_with("end") => format!("unexpected closing tag `{tag}`"),
+        tag => format!("unknown node `{tag}`"),
+    };
+    Err(nom::Err::Failure(ErrorContext::new(msg, i)))
+}
+
 #[derive(Debug, PartialEq)]
 pub struct When<'a> {
     pub ws: Ws,
@@ -175,17 +250,20 @@ pub struct When<'a> {
 }
 
 impl<'a> When<'a> {
-    fn r#match(i: &'a str, s: &State<'_>) -> ParseResult<'a, WithSpan<'a, Self>> {
+    fn r#else(i: &'a str, s: &State<'_>) -> ParseResult<'a, WithSpan<'a, Self>> {
         let start = i;
         let mut p = tuple((
             |i| s.tag_block_start(i),
             opt(Whitespace::parse),
             ws(keyword("else")),
-            cut(tuple((
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-                cut(|i| Node::many(i, s)),
-            ))),
+            cut_node(
+                Some("match-else"),
+                tuple((
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                    cut_node(Some("match-else"), |i| Node::many(i, s)),
+                )),
+            ),
         ));
         let (i, (_, pws, _, (nws, _, nodes))) = p(i)?;
         Ok((
@@ -204,18 +282,55 @@ impl<'a> When<'a> {
     #[allow(clippy::self_named_constructors)]
     fn when(i: &'a str, s: &State<'_>) -> ParseResult<'a, WithSpan<'a, Self>> {
         let start = i;
+        let endwhen = map(
+            consumed(ws(pair(
+                delimited(
+                    |i| s.tag_block_start(i),
+                    opt(Whitespace::parse),
+                    ws(keyword("endwhen")),
+                ),
+                cut_node(
+                    Some("match-endwhen"),
+                    tuple((
+                        opt(Whitespace::parse),
+                        |i| s.tag_block_end(i),
+                        many0(value((), ws(|i| Comment::parse(i, s)))),
+                    )),
+                ),
+            ))),
+            |(span, (pws, _))| {
+                // A comment node is used to pass the whitespace suppressing information to the
+                // generator. This way we don't have to fix up the next `when` node or the closing
+                // `endmatch`. Any whitespaces after `endwhen` are to be suppressed. Actually, they
+                // don't wind up in the AST anyway.
+                Node::Comment(WithSpan::new(
+                    Comment {
+                        ws: Ws(pws, Some(Whitespace::Suppress)),
+                        content: "",
+                    },
+                    span,
+                ))
+            },
+        );
         let mut p = tuple((
             |i| s.tag_block_start(i),
             opt(Whitespace::parse),
             ws(keyword("when")),
-            cut(tuple((
-                separated_list1(char('|'), ws(|i| Target::parse(i, s))),
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-                cut(|i| Node::many(i, s)),
-            ))),
+            cut_node(
+                Some("match-when"),
+                tuple((
+                    separated_list1(char('|'), ws(|i| Target::parse(i, s))),
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                    cut_node(Some("match-when"), |i| Node::many(i, s)),
+                    opt(endwhen),
+                )),
+            ),
         ));
-        let (i, (_, pws, _, (target, nws, _, nodes))) = p(i)?;
+        let (i, (_, pws, _, (target, nws, _, mut nodes, endwhen))) = p(i)?;
+        if let Some(endwhen) = endwhen {
+            nodes.push(endwhen);
+        }
         Ok((
             i,
             WithSpan::new(
@@ -247,12 +362,12 @@ impl<'a> Cond<'a> {
                 preceded(ws(keyword("else")), opt(|i| CondTest::parse(i, s))),
                 preceded(
                     ws(keyword("elif")),
-                    cut(map(|i| CondTest::parse_cond(i, s), Some)),
+                    cut_node(Some("if-elif"), map(|i| CondTest::parse_cond(i, s), Some)),
                 ),
             )),
             opt(Whitespace::parse),
-            cut(|i| s.tag_block_end(i)),
-            cut(|i| Node::many(i, s)),
+            cut_node(Some("if"), |i| s.tag_block_end(i)),
+            cut_node(Some("if"), |i| Node::many(i, s)),
         ))(i)?;
         Ok((
             i,
@@ -277,7 +392,10 @@ pub struct CondTest<'a> {
 
 impl<'a> CondTest<'a> {
     fn parse(i: &'a str, s: &State<'_>) -> ParseResult<'a, Self> {
-        preceded(ws(keyword("if")), cut(|i| Self::parse_cond(i, s)))(i)
+        preceded(
+            ws(keyword("if")),
+            cut_node(Some("if"), |i| Self::parse_cond(i, s)),
+        )(i)
     }
 
     fn parse_cond(i: &'a str, s: &State<'_>) -> ParseResult<'a, Self> {
@@ -310,12 +428,33 @@ pub enum Whitespace {
 
 impl Whitespace {
     fn parse(i: &str) -> ParseResult<'_, Self> {
-        alt((
-            value(Self::Preserve, char('+')),
-            value(Self::Suppress, char('-')),
-            value(Self::Minimize, char('~')),
-        ))(i)
+        map_opt(anychar, Self::parse_char)(i)
     }
+
+    fn parse_char(c: char) -> Option<Self> {
+        match c {
+            '+' => Some(Self::Preserve),
+            '-' => Some(Self::Suppress),
+            '~' => Some(Self::Minimize),
+            _ => None,
+        }
+    }
+}
+
+fn check_block_start<'a>(
+    i: &'a str,
+    start: &'a str,
+    s: &State<'_>,
+    node: &str,
+    expected: &str,
+) -> ParseResult<'a> {
+    if i.is_empty() {
+        return Err(nom::Err::Failure(ErrorContext::new(
+            format!("expected `{expected}` to terminate `{node}` node, found nothing"),
+            start,
+        )));
+    }
+    s.tag_block_start(i)
 }
 
 #[derive(Debug, PartialEq)]
@@ -342,53 +481,72 @@ impl<'a> Loop<'a> {
         let start = i;
         let if_cond = preceded(
             ws(keyword("if")),
-            cut(ws(|i| Expr::parse(i, s.level.get()))),
+            cut_node(Some("for-if"), ws(|i| Expr::parse(i, s.level.get()))),
         );
 
         let else_block = |i| {
             let mut p = preceded(
                 ws(keyword("else")),
-                cut(tuple((
-                    opt(Whitespace::parse),
-                    delimited(
-                        |i| s.tag_block_end(i),
-                        |i| Node::many(i, s),
-                        |i| s.tag_block_start(i),
-                    ),
-                    opt(Whitespace::parse),
-                ))),
+                cut_node(
+                    Some("for-else"),
+                    tuple((
+                        opt(Whitespace::parse),
+                        delimited(
+                            |i| s.tag_block_end(i),
+                            |i| Node::many(i, s),
+                            |i| s.tag_block_start(i),
+                        ),
+                        opt(Whitespace::parse),
+                    )),
+                ),
             );
             let (i, (pws, nodes, nws)) = p(i)?;
             Ok((i, (pws, nodes, nws)))
         };
 
-        let mut p = tuple((
-            opt(Whitespace::parse),
-            ws(keyword("for")),
-            cut(tuple((
-                ws(|i| Target::parse(i, s)),
-                ws(keyword("in")),
-                cut(tuple((
-                    ws(|i| Expr::parse(i, s.level.get())),
-                    opt(if_cond),
-                    opt(Whitespace::parse),
-                    |i| s.tag_block_end(i),
-                    cut(tuple((
-                        |i| content(i, s),
-                        cut(tuple((
-                            |i| s.tag_block_start(i),
+        let body_and_end = |i| {
+            let (i, (body, (_, pws, else_block, _, nws))) = cut_node(
+                Some("for"),
+                tuple((
+                    |i| content(i, s),
+                    cut_node(
+                        Some("for"),
+                        tuple((
+                            |i| check_block_start(i, start, s, "for", "endfor"),
                             opt(Whitespace::parse),
                             opt(else_block),
                             end_node("for", "endfor"),
                             opt(Whitespace::parse),
-                        ))),
-                    ))),
-                ))),
-            ))),
+                        )),
+                    ),
+                )),
+            )(i)?;
+            Ok((i, (body, pws, else_block, nws)))
+        };
+
+        let mut p = tuple((
+            opt(Whitespace::parse),
+            ws(keyword("for")),
+            cut_node(
+                Some("for"),
+                tuple((
+                    ws(|i| Target::parse(i, s)),
+                    ws(keyword("in")),
+                    cut_node(
+                        Some("for"),
+                        tuple((
+                            ws(|i| Expr::parse(i, s.level.get())),
+                            opt(if_cond),
+                            opt(Whitespace::parse),
+                            |i| s.tag_block_end(i),
+                            body_and_end,
+                        )),
+                    ),
+                )),
+            ),
         ));
-        let (i, (pws1, _, (var, _, (iter, cond, nws1, _, (body, (_, pws2, else_block, _, nws2)))))) =
-            p(i)?;
-        let (nws3, else_block, pws3) = else_block.unwrap_or_default();
+        let (i, (pws1, _, (var, _, (iter, cond, nws1, _, (body, pws2, else_block, nws2))))) = p(i)?;
+        let (nws3, else_nodes, pws3) = else_block.unwrap_or_default();
         Ok((
             i,
             WithSpan::new(
@@ -399,7 +557,7 @@ impl<'a> Loop<'a> {
                     cond,
                     body,
                     ws2: Ws(pws2, nws3),
-                    else_nodes: else_block,
+                    else_nodes,
                     ws3: Ws(pws3, nws2),
                 },
                 start,
@@ -431,12 +589,15 @@ impl<'a> Macro<'a> {
         let mut start = tuple((
             opt(Whitespace::parse),
             ws(keyword("macro")),
-            cut(tuple((
-                ws(identifier),
-                opt(ws(parameters)),
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-            ))),
+            cut_node(
+                Some("macro"),
+                tuple((
+                    ws(identifier),
+                    opt(ws(parameters)),
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                )),
+            ),
         ));
         let (j, (pws1, _, (name, params, nws1, _))) = start(i)?;
         if is_rust_keyword(name) {
@@ -446,21 +607,30 @@ impl<'a> Macro<'a> {
             )));
         }
 
-        let mut end = cut(tuple((
-            |i| Node::many(i, s),
-            cut(tuple((
-                |i| s.tag_block_start(i),
-                opt(Whitespace::parse),
-                end_node("macro", "endmacro"),
-                cut(preceded(
-                    opt(|before| {
-                        let (after, end_name) = ws(identifier)(before)?;
-                        check_end_name(before, after, name, end_name, "macro")
-                    }),
-                    opt(Whitespace::parse),
-                )),
-            ))),
-        )));
+        let mut end = cut_node(
+            Some("macro"),
+            tuple((
+                |i| Node::many(i, s),
+                cut_node(
+                    Some("macro"),
+                    tuple((
+                        |i| check_block_start(i, start_s, s, "macro", "endmacro"),
+                        opt(Whitespace::parse),
+                        end_node("macro", "endmacro"),
+                        cut_node(
+                            Some("macro"),
+                            preceded(
+                                opt(|before| {
+                                    let (after, end_name) = ws(identifier)(before)?;
+                                    check_end_name(before, after, name, end_name, "macro")
+                                }),
+                                opt(Whitespace::parse),
+                            ),
+                        ),
+                    )),
+                ),
+            )),
+        );
         let (i, (contents, (_, pws2, _, nws2))) = end(j)?;
 
         Ok((
@@ -494,14 +664,19 @@ impl<'a> FilterBlock<'a> {
         let mut start = tuple((
             opt(Whitespace::parse),
             ws(keyword("filter")),
-            cut(tuple((
-                ws(identifier),
-                opt(|i| Expr::arguments(i, s.level.get(), false)),
-                many0(|i| filter(i, &mut level).map(|(j, (name, params))| (j, (name, params, i)))),
-                ws(|i| Ok((i, ()))),
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-            ))),
+            cut_node(
+                Some("filter"),
+                tuple((
+                    ws(identifier),
+                    opt(|i| Expr::arguments(i, s.level.get(), false)),
+                    many0(|i| {
+                        filter(i, &mut level).map(|(j, (name, params))| (j, (name, params, i)))
+                    }),
+                    ws(|i| Ok((i, ()))),
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                )),
+            ),
         ));
         let (i, (pws1, _, (filter_name, params, extra_filters, _, nws1, _))) = start(i)?;
 
@@ -522,15 +697,21 @@ impl<'a> FilterBlock<'a> {
             };
         }
 
-        let mut end = cut(tuple((
-            |i| Node::many(i, s),
-            cut(tuple((
-                |i| s.tag_block_start(i),
-                opt(Whitespace::parse),
-                end_node("filter", "endfilter"),
-                opt(Whitespace::parse),
-            ))),
-        )));
+        let mut end = cut_node(
+            Some("filter"),
+            tuple((
+                |i| Node::many(i, s),
+                cut_node(
+                    Some("filter"),
+                    tuple((
+                        |i| check_block_start(i, start_s, s, "filter", "endfilter"),
+                        opt(Whitespace::parse),
+                        end_node("filter", "endfilter"),
+                        opt(Whitespace::parse),
+                    )),
+                ),
+            )),
+        );
         let (i, (nodes, (_, pws2, _, nws2))) = end(i)?;
 
         Ok((
@@ -561,11 +742,14 @@ impl<'a> Import<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(keyword("import")),
-            cut(tuple((
-                ws(str_lit_without_prefix),
-                ws(keyword("as")),
-                cut(pair(ws(identifier), opt(Whitespace::parse))),
-            ))),
+            cut_node(
+                Some("import"),
+                tuple((
+                    ws(str_lit_without_prefix),
+                    ws(keyword("as")),
+                    cut_node(Some("import"), pair(ws(identifier), opt(Whitespace::parse))),
+                )),
+            ),
         ));
         let (i, (pws, _, (path, _, (scope, nws)))) = p(i)?;
         Ok((
@@ -596,12 +780,15 @@ impl<'a> Call<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(keyword("call")),
-            cut(tuple((
-                opt(tuple((ws(identifier), ws(tag("::"))))),
-                ws(identifier),
-                opt(ws(|nested| Expr::arguments(nested, s.level.get(), true))),
-                opt(Whitespace::parse),
-            ))),
+            cut_node(
+                Some("call"),
+                tuple((
+                    opt(tuple((ws(identifier), ws(tag("::"))))),
+                    ws(identifier),
+                    opt(ws(|nested| Expr::arguments(nested, s.level.get(), true))),
+                    opt(Whitespace::parse),
+                )),
+            ),
         ));
         let (i, (pws, _, (scope, name, args, nws))) = p(i)?;
         let scope = scope.map(|(scope, _)| scope);
@@ -635,24 +822,38 @@ impl<'a> Match<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(keyword("match")),
-            cut(tuple((
-                ws(|i| Expr::parse(i, s.level.get())),
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-                cut(tuple((
-                    ws(many0(ws(value((), |i| Comment::parse(i, s))))),
-                    many0(|i| When::when(i, s)),
-                    cut(tuple((
-                        opt(|i| When::r#match(i, s)),
-                        cut(tuple((
-                            ws(|i| s.tag_block_start(i)),
-                            opt(Whitespace::parse),
-                            end_node("match", "endmatch"),
-                            opt(Whitespace::parse),
-                        ))),
-                    ))),
-                ))),
-            ))),
+            cut_node(
+                Some("match"),
+                tuple((
+                    ws(|i| Expr::parse(i, s.level.get())),
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                    cut_node(
+                        Some("match"),
+                        tuple((
+                            ws(many0(ws(value((), |i| Comment::parse(i, s))))),
+                            many0(|i| When::when(i, s)),
+                            cut_node(
+                                Some("match"),
+                                tuple((
+                                    opt(|i| When::r#else(i, s)),
+                                    cut_node(
+                                        Some("match"),
+                                        tuple((
+                                            ws(|i| {
+                                                check_block_start(i, start, s, "match", "endmatch")
+                                            }),
+                                            opt(Whitespace::parse),
+                                            end_node("match", "endmatch"),
+                                            opt(Whitespace::parse),
+                                        )),
+                                    ),
+                                )),
+                            ),
+                        )),
+                    ),
+                )),
+            ),
         ));
         let (i, (pws1, _, (expr, nws1, _, (_, mut arms, (else_arm, (_, pws2, _, nws2)))))) = p(i)?;
 
@@ -695,27 +896,39 @@ impl<'a> BlockDef<'a> {
         let mut start = tuple((
             opt(Whitespace::parse),
             ws(keyword("block")),
-            cut(tuple((ws(identifier), opt(Whitespace::parse), |i| {
-                s.tag_block_end(i)
-            }))),
+            cut_node(
+                Some("block"),
+                tuple((ws(identifier), opt(Whitespace::parse), |i| {
+                    s.tag_block_end(i)
+                })),
+            ),
         ));
         let (i, (pws1, _, (name, nws1, _))) = start(i)?;
 
-        let mut end = cut(tuple((
-            |i| Node::many(i, s),
-            cut(tuple((
-                |i| s.tag_block_start(i),
-                opt(Whitespace::parse),
-                end_node("block", "endblock"),
-                cut(tuple((
-                    opt(|before| {
-                        let (after, end_name) = ws(identifier)(before)?;
-                        check_end_name(before, after, name, end_name, "block")
-                    }),
-                    opt(Whitespace::parse),
-                ))),
-            ))),
-        )));
+        let mut end = cut_node(
+            Some("block"),
+            tuple((
+                |i| Node::many(i, s),
+                cut_node(
+                    Some("block"),
+                    tuple((
+                        |i| check_block_start(i, start_s, s, "block", "endblock"),
+                        opt(Whitespace::parse),
+                        end_node("block", "endblock"),
+                        cut_node(
+                            Some("block"),
+                            tuple((
+                                opt(|before| {
+                                    let (after, end_name) = ws(identifier)(before)?;
+                                    check_end_name(before, after, name, end_name, "block")
+                                }),
+                                opt(Whitespace::parse),
+                            )),
+                        ),
+                    )),
+                ),
+            )),
+        );
         let (i, (nodes, (_, pws2, _, (_, nws2)))) = end(i)?;
 
         Ok((
@@ -821,11 +1034,14 @@ impl<'a> Raw<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(keyword("raw")),
-            cut(tuple((
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-                consumed(skip_till(Splitter1::new(s.syntax.block_start), endraw)),
-            ))),
+            cut_node(
+                Some("raw"),
+                tuple((
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                    consumed(skip_till(Splitter1::new(s.syntax.block_start), endraw)),
+                )),
+            ),
         ));
 
         let (_, (pws1, _, (nws1, _, (contents, (i, (_, pws2, _, nws2, _)))))) = p(i)?;
@@ -849,14 +1065,17 @@ impl<'a> Let<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(alt((keyword("let"), keyword("set")))),
-            cut(tuple((
-                ws(|i| Target::parse(i, s)),
-                opt(preceded(
-                    ws(char('=')),
-                    ws(|i| Expr::parse(i, s.level.get())),
+            cut_node(
+                Some("let"),
+                tuple((
+                    ws(|i| Target::parse(i, s)),
+                    opt(preceded(
+                        ws(char('=')),
+                        ws(|i| Expr::parse(i, s.level.get())),
+                    )),
+                    opt(Whitespace::parse),
                 )),
-                opt(Whitespace::parse),
-            ))),
+            ),
         ));
         let (i, (pws, _, (var, val, nws))) = p(i)?;
 
@@ -886,20 +1105,29 @@ impl<'a> If<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             |i| CondTest::parse(i, s),
-            cut(tuple((
-                opt(Whitespace::parse),
-                |i| s.tag_block_end(i),
-                cut(tuple((
-                    |i| Node::many(i, s),
-                    many0(|i| Cond::parse(i, s)),
-                    cut(tuple((
-                        |i| s.tag_block_start(i),
-                        opt(Whitespace::parse),
-                        end_node("if", "endif"),
-                        opt(Whitespace::parse),
-                    ))),
-                ))),
-            ))),
+            cut_node(
+                Some("if"),
+                tuple((
+                    opt(Whitespace::parse),
+                    |i| s.tag_block_end(i),
+                    cut_node(
+                        Some("if"),
+                        tuple((
+                            |i| Node::many(i, s),
+                            many0(|i| Cond::parse(i, s)),
+                            cut_node(
+                                Some("if"),
+                                tuple((
+                                    |i| check_block_start(i, start, s, "if", "endif"),
+                                    opt(Whitespace::parse),
+                                    end_node("if", "endif"),
+                                    opt(Whitespace::parse),
+                                )),
+                            ),
+                        )),
+                    ),
+                )),
+            ),
         ));
 
         let (i, (pws1, cond, (nws1, _, (nodes, elifs, (_, pws2, _, nws2))))) = p(i)?;
@@ -938,7 +1166,10 @@ impl<'a> Include<'a> {
         let mut p = tuple((
             opt(Whitespace::parse),
             ws(keyword("include")),
-            cut(pair(ws(str_lit_without_prefix), opt(Whitespace::parse))),
+            cut_node(
+                Some("include"),
+                pair(ws(str_lit_without_prefix), opt(Whitespace::parse)),
+            ),
         ));
         let (i, (pws, _, (path, nws))) = p(i)?;
         Ok((
@@ -966,7 +1197,10 @@ impl<'a> Extends<'a> {
         let (i, (pws, _, (path, nws))) = tuple((
             opt(Whitespace::parse),
             ws(keyword("extends")),
-            cut(pair(ws(str_lit_without_prefix), opt(Whitespace::parse))),
+            cut_node(
+                Some("extends"),
+                pair(ws(str_lit_without_prefix), opt(Whitespace::parse)),
+            ),
         ))(i)?;
         match (pws, nws) {
             (None, None) => Ok((i, WithSpan::new(Self { path }, start))),
@@ -999,12 +1233,12 @@ impl<'a> Comment<'a> {
             ))(i)
         }
 
-        fn content<'a>(mut i: &'a str, s: &State<'_>) -> ParseResult<'a, ()> {
+        fn content<'a>(mut i: &'a str, s: &State<'_>) -> ParseResult<'a> {
             let mut depth = 0usize;
+            let start = i;
             loop {
-                let start = i;
                 let splitter = Splitter2::new(s.syntax.comment_start, s.syntax.comment_end);
-                let (_, tag) = opt(skip_till(splitter, |i| tag(i, s)))(i)?;
+                let (k, tag) = opt(skip_till(splitter, |i| tag(i, s)))(i)?;
                 let Some((j, tag)) = tag else {
                     return Err(
                         ErrorContext::unclosed("comment", s.syntax.comment_end, start).into(),
@@ -1022,7 +1256,7 @@ impl<'a> Comment<'a> {
                     },
                     Tag::Close => match depth.checked_sub(1) {
                         Some(new_depth) => depth = new_depth,
-                        None => return Ok((j, ())),
+                        None => return Ok((j, &start[..start.len() - k.len()])),
                     },
                 }
                 i = j;
@@ -1030,31 +1264,28 @@ impl<'a> Comment<'a> {
         }
 
         let start = i;
-        let (i, (pws, content)) = pair(
-            preceded(|i| s.tag_comment_start(i), opt(Whitespace::parse)),
-            recognize(cut(|i| content(i, s))),
+        let (i, content) = preceded(
+            |i| s.tag_comment_start(i),
+            cut_node(Some("comment"), |i| content(i, s)),
         )(i)?;
 
-        let mut nws = None;
-        if let Some(content) = content.strip_suffix(s.syntax.comment_end) {
-            nws = match content.chars().last() {
-                Some('-') => Some(Whitespace::Suppress),
-                Some('+') => Some(Whitespace::Preserve),
-                Some('~') => Some(Whitespace::Minimize),
-                _ => None,
-            }
-        };
-
-        Ok((
-            i,
-            WithSpan::new(
-                Self {
-                    ws: Ws(pws, nws),
-                    content,
-                },
+        let mut ws = Ws(None, None);
+        if content.len() == 1 && matches!(content, "-" | "+" | "~") {
+            return Err(nom::Err::Failure(ErrorContext::new(
+                format!(
+                    "ambiguous whitespace stripping\n\
+                     use `{}{content} {content}{}` to apply the same whitespace stripping on both \
+                     sides",
+                    s.syntax.comment_start, s.syntax.comment_end,
+                ),
                 start,
-            ),
-        ))
+            )));
+        } else if content.len() >= 2 {
+            ws.0 = Whitespace::parse_char(content.chars().next().unwrap_or_default());
+            ws.1 = Whitespace::parse_char(content.chars().next_back().unwrap_or_default());
+        }
+
+        Ok((i, WithSpan::new(Self { ws, content }, start)))
     }
 }
 
