@@ -721,6 +721,7 @@ fn test_odd_calls() {
 
 #[test]
 fn test_parse_comments() {
+    #[track_caller]
     fn one_comment_ws(source: &str, ws: Ws) {
         let s = &Syntax::default();
         let mut nodes = Ast::from_str(source, None, s).unwrap().nodes;
@@ -913,10 +914,10 @@ fn test_parse_tuple() {
 fn test_missing_space_after_kw() {
     let syntax = Syntax::default();
     let err = Ast::from_str("{%leta=b%}", None, &syntax).unwrap_err();
-    assert!(matches!(
-        &*err.to_string(),
-        "failed to parse template source near offset 0",
-    ));
+    assert_eq!(
+        err.to_string(),
+        "unknown node `leta`\nfailed to parse template source near offset 2",
+    );
 }
 
 #[test]
